
/*
 *  Sun Nov 14 18:17:40 EET 2004
 *
 *  Commands:
 *  42                     address radio with CIV address 0x42.
 *  ic735                  act specially for ic735 until again addressed.
 *  3699[.NNN]             kHz frequency (3 or more digits, decimal optional).
 *  mN                     memory N (1 or two digits).
 *  usb/lsb/cw/am/fm/etc   set mode of adressed radio.
 *  subNN.N                ctcss Hz
 *  txoff/txon             at least 706 rejects this command.
 *  mox/vox                manual/vox
 *  sql?                   query sql state
 *  smeter?                query signal strength
 *  ?                      ask a few simple questions, try with Addr=00
 *
 *  example command: 42 ic735 lsb 3699.000
 */

#include "standard.h"
#include "hamdr_cfg.h"

static int
	Us  = 0xE0,    /* controller normal address */
	All = 0x00,    /* everyone hear this */
	Addr;          /* current 'to' (default All) */

static int
	ic735,
	idle,
	pace = 100,    /* 100 msec */
	timo = 10000;  /* 10 seconds retrying */

static struct sendq {
	long timo;
	struct sendq *next;
	byte len;
	byte buf[32];
} *pending;

static byte cividx;
static byte civbuf[32];

static struct {
	char name[4];
	byte value;
} modes[] = {
	{ "LSB", 0x00 },
	{ "USB", 0x01 },
	{ "AM",  0x02 },
	{ "CW",  0x03 },
	{ "FSK", 0x04 },
	{ "FM",  0x05 },
	{ "WFM", 0x06 },
};

static char *smeter_table[] = {
	"S0", "S1", "S2", "S3", "S4", "S5", "S6", "S7", "S8", "S9",
	"S9+10", "S9+20", "S9+30", "S9+40", "S9+50", "S9+60",
};

struct sendq *
alloc_sendq(int len)
{
	struct sendq *q;

	if (len > sizeof(q->buf))
		doh("Too large civ qelem");

	q = ;

	q->next = NULL;
	q->timo = 0;
	q->len  = len;

	return (q);
}

static void
civ_write(struct sendq *q)
{
	int i;

	/*
	 *  No flow control at all, about 1000 bytes per second (XXX)
	 *  blocking not considered.
	 */
	idle = FALSE;

	SYS_PutChar(0xFE);
	SYS_PutChar(0xFE);
	for (i = 0; i < q->len; ++i)
		SYS_PutChar(q->buf[i]);
	SYS_PutChar(0xFD);
}

static void
civ_command(byte to, byte cmd, void *data, int dlen)
{
	struct sendq *q, **qq;

	q = alloc_sendq(3 + dlen);
	q->buf[0] = to;
	q->buf[1] = Us;
	q->buf[2] = cmd;

	if (dlen)
		bcopy(data, &q->buf[3], dlen);

	qq = &pending;
	while (*qq)
		qq = &(*qq)->next;
	*qq = q;

	/* sloppy collision avoider */

	if (idle)
		civ_write(q);
}

static void
civ_simple1(byte to, byte cmd)
{
	civ_command(to, cmd, NULL, 0);
}

static void
civ_simple2(byte to, byte cmd, byte sub)
{
	civ_command(to, cmd, &sub, 1);
}

static void
civ_simple3(byte to, byte cmd, byte sub, byte b0)
{
	byte data[2];

	data[0] = sub;
	data[1] = b0;
	civ_command(to, cmd, data, sizeof(data));
}

static void
civ_simple4(byte to, byte cmd, byte sub, byte b0, byte b1)
{
	byte data[3];

	data[0] = sub;
	data[1] = b0;
	data[2] = b1;
	civ_command(to, cmd, data, sizeof(data));
}

static void
unpend(byte *buf, int len)
{
	struct sendq *q, **qq;

	for (qq = &pending; q = *qq; qq = &q->next)
		if (q->len == len && !bcmp(q->buf, buf, len))
			break;

	if (q) {
		*qq = q->next;
		free_sendq(q);
	} else
		civ_error("Unsent echo ?");
}

static void
itohz(long hz, byte *bcd, int n)
{
	do {
		*bcd = hz % 10;
		hz /= 10;
		*bcd |= (hz % 10) << 4;
		hz /= 10;
		bcd++;
	} while (--n);

	if (hz)
		civ_error("Frequency overflow");
}

static long
hztoi(byte *bcd, int n)
{
	long hz = 0;

	while (--n >= 0) {
		hz *= 10;
		hz += (bcd[n] & 0xF0) >> 4;
		hz *= 10;
		hz += (bcd[n] & 0x0F) >> 0;
	}
	return (hz);
}

static char *
modetoa(byte mode)
{
	int i;

	for (i = 0; i < __SIZEOF(modes); ++i)
		if (mode == modes[i].value)
			return (modes[i].name);

	return (NULL);
}

static int
atomode(char *name)
{
	int i;

	for (i = 0; i < __SIZEOF(modes); ++i)
		if (!strcasecmp(modes[i].name, name))
			return (modes[i].value);

	return (-1);
}

static void
civ_input(byte *buf, int len)       /* to from cmd [data] */
{
	byte *data;
	long hz;
	int val;
	char *s, *width;

	if (buf[1] == Us) {     /* echo */
		unpend(buf, len);
		return;
	}
	if (buf[0] != All && buf[0] != Us) /* not for us */
		return;

	data = &buf[3];
	len -= 3;           /* data[len] is now the optional data */

	switch (buf[2]) {
	default:
		break;

		/* XXX peel head of pending here, not in echo */

	case 0xFA:
		fprintf(stderr, "%02X nak\n", buf[1]);
		break;
	case 0xFB:
		fprintf(stderr, "%02X ack\n", buf[1]);
		break;

	/* frequency. 4 or 5 digits */

	case 0x00:
		if (len < 1)
			return;

		hz = hztoi(data, len);
		fprintf(stderr, "%02X %lu.%03lu\n", from, hz / 1000, hz % 1000);
		break;

	/* mode */

	case 0x01:
		if (len < 1)
			return;

		s = modetoa(data[0]);
		if (len > 1 && data[1] == 0x02 && s && !strcmp(s, "FM"))
			width = "W";
		else
			width = "";
		if (s)
			fprintf(stderr, "%02X %s%s\n", from, width, s);
		else
			fprintf(stderr, "%02X mode%d\n", from, data[0]);
		break;

	case 0x1B:
		if (len < 1)
			return;

		switch (data[0]) {
		case 0x00:              /* TONE 1B 00 NN NN */
			if (len < 3)
				return;

			val = (data[1] >> 4) * 1000
			    + (data[1] & 0x0F) * 100
			    + (data[2] >> 4) * 10
			    + (data[2] & 0x0F);
			fprintf(stderr, "%02X sub%d.%d\n", from, val / 10, val % 10);
			break;
		}
		break;

	case 0x15:
		if (len < 1)
			return;

		switch (data[0]) {

		case 0x01:              /* SQL 15 01 00/01 */
			if (len < 2)
				return;
			fprintf(stderr, "%02X sql %s\n", from, data[1] ? "open" : "closed");
			break;

		case 0x02:              /* S-meter 15 02 02 55 */
			if (len < 3)
				return;
			val = (data[1] & 0x0F) * 100
			    + (data[2] >> 4) * 10
			    + (data[2] & 0x0F);
			if (val < 0)
				val = 0;
			if (val > 255)
				val = 255;
			val = val * 16 / 256;
			fprintf(stderr, "%02X signal %s\n", from, smeter_table[val]);
			break;
		}
		break;

	}
}

static void
civ_character(byte c)
{
	if (c == 0xFC) {                /* someone saw a collision */
		civ_error("Collision");
		goto rewind;
	}
	if (c != 0xFE && cividx < 2) {  /* always starts with FE FE */
		civ_error("Garbage");
		goto rewind;
	}
	if (cividx >= sizeof(civbuf)) {
		civ_error("Overflow");
		goto rewind;
	}
	civbuf[cividx++] = c;

	if (c != 0xFD)
		return;         /* more to come */

	if (cividx >= 2 + 3 + 1)
		civ_input(civbuf + 2, cividx - 2 - 1); /* strip FE FE and FD */

rewind:
	cividx = 0;
}

static void
user_command(char *cmd)
{
	int i;

	/* 'to' ? */

	if (strlen(cmd) == 2 && isxdigit(cmd[0]) && isxdigit(cmd[1])) {
		sscanf(cmd, "%2x", &Addr);
		ic735 = 0;
		return;
	}
	if (!strcasecmp(cmd, "ic735")) {
		ic735 = 1;
		return;
	}

	/* set memory ? */

	if (!strncasecmp(cmd, "m", 1) && strlen(cmd) > 1 && isdigit(cmd[1])) {
		i = atoi(&cmd[1]);
		i = ((i / 10 % 10) << 4) | (i % 10);  /* BCD memory number */

		civ_simple2(Addr, 0x08, i);
		civ_simple1(Addr, 0x08);        /* memory mode */
		return;
	}
	/* set frequency ? */

	if (isdigit(*cmd) && strlen(cmd) >= 3) {
		unsigned long freq = atof(cmd) * 1000;
		int n = ic735 ? 4 : 5; /* frequency format shorter for 731/735 */
		byte buf[5];

		itohz(freq, buf, n);
		civ_command(Addr, 0x05, buf, n);
		civ_simple1(Addr, 0x07);          /* VFO mode */
		return;
	}
	/* tx control ? */

	if (!strcasecmp(cmd, "txon")) {
		civ_simple3(Addr, 0x1C, 0x00, 0x01);
		return;
	}
	if (!strcasecmp(cmd, "txoff")) {
		civ_simple3(Addr, 0x1C, 0x00, 0x00);
		return;
	}
	if (!strncasecmp(cmd, "sub", 3)) {
		i = atof(cmd + 3) * 10; /* 118.8 to 1188 */

		i = ((i / 1000 % 10) << 12)
		  | ((i /  100 % 10) <<  8)
		  | ((i /   10 % 10) <<  4)
		  | ((i /    1 % 10) <<  0);
		civ_simple4(Addr, 0x1B, 0x00, i >> 8, i);
		return;
	}

	/* sql/smeter ? */

	if (!strcasecmp(cmd, "sql?")) {
		civ_simple2(Addr, 0x15, 0x01);
		return;
	}
	if (!strcasecmp(cmd, "smeter?")) {
		civ_simple2(Addr, 0x15, 0x02);
		return;
	}
	if (!strcasecmp(cmd, "?")) {
		civ_simple1(Addr, 0x03);       /* read freq */
		civ_simple1(Addr, 0x04);       /* read mode */
		civ_simple1(Addr, 0x0C);       /* read duplex */
		civ_simple2(Addr, 0x15, 0x01); /* read sql state */
		civ_simple2(Addr, 0x15, 0x02); /* read s-meter */
		civ_simple2(Addr, 0x1B, 0x00); /* read tone frequency */
		return;
	}
	/* vox control ? */

	if (!strcasecmp(cmd, "mox")) {
		civ_simple3(Addr, 0x16, 0x46, 0x00);
		return;
	}
	if (!strcasecmp(cmd, "vox")) {
		civ_simple3(Addr, 0x16, 0x46, 0x01);
		return;
	}

	/* set mode ? */

	if ((i = atomode(cmd)) != -1) {
		civ_simple2(Addr, 0x06, i);
		return;
	}

	/* hmph. dunno what that was */

	civ_error("Invalid command");
}

void
civ_link(void)
{
	SYS_SetBaud
	idle = FALSE;

	for (;;) {
		if (pending && SYS_TestTimeL(pending->timo)) {
			struct sendq *q;

			while (q = pending) {
				pending = q->next;
				free_sendq(q);
			}
			civ_error("no echo, sendq purged. sorry");
		}
		if (pending) {
			tvp = &tv;
			tvp->tv_sec = 0;
			tvp->tv_usec = pace * 1000;
		} else
			tvp = NULL;

		nready = select(nfd, &rset, NULL, NULL, tvp);

		if (nready == 0) {
			idle = TRUE;
			if (pending) {
				pending->timo += pace;
				civ_write(pending);
			}
			continue;
		}
		while (SYS_ChkChar()) {
			idle = FALSE;
			civ_character(SYS_GetChar());
		}
		if (user input) {
			user_command(cmd line);
		}
	}
}

